from .binarisation_methods import*
from .Declustering import*
from .spatial_methods import*
from .utils import*
from .validation_tools import*

print("init file executed")
print("###########################################")
print(" Hi! thank you for downloaded this package!")
print("If there is a problem in the code or a suggestion, please email me at maxime.hoareau@ens-lyon.fr")
print("For further information see the realted article:")
print("type help(func), for a given func to see all possibilities")
print("###########################################")
print("USEFUL functions for users:")
print(">>>CurrRange (main)")
print(">>>runoverfile (main over file)")
print(">>>BoyceIndexTresh (binarisation)")
print(">>>PseudoRange (reference range reconstruction)")
print("############################################")